import filecmp
import os.path
import shutil
import unittest
from pathlib import PurePath

import tensorflow as tf
from tempfile import TemporaryDirectory

from ..core import (
    CustomNamedTemporaryFile,
    encode_dir_to_file,
    decode_file_to_dir,
)

base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
folder_path = os.path.join(base_dir, "tests")


class TestEncodeDecode(unittest.TestCase):
    def setUp(self) -> None:
        # setup file structure

        model: tf.keras.Sequential = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(512, activation="relu", input_shape=(111,)),
                tf.keras.layers.Dropout(0.2),
                tf.keras.layers.Dense(4),
            ]
        )

        # TODO find multiclass classification loss and metrics etc.
        model.compile(
            optimizer="adam",
            loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
            metrics=[tf.keras.metrics.SparseCategoricalAccuracy()],
        )

        self.root = TemporaryDirectory()
        tf.keras.models.save_model(model, self.root.name, save_format="tf")

    def test_dir_to_file_to_dir(self):

        a = self.root.name
        with CustomNamedTemporaryFile() as temp:
            with open(temp.name, "wb") as write_f:
                encode_dir_to_file(PurePath(self.root.name), file=write_f)
            with open(temp.name, "rb") as read_f:
                b = decode_file_to_dir(read_f)

        comparer = filecmp.dircmp(a, b)
        print(comparer.report())

        # keeping this in as this is a primary use case that must work - error here means we have issues.
        tf.keras.models.load_model(b)

        self.assertEqual(
            comparer.diff_files,
            [],
            "Some files are different after decoding...",
        )
        shutil.rmtree(b)


if __name__ == "__main__":
    unittest.main()



from unittest import TestCase

import numpy as np
import pandas as pd

from ..dataset_management import dataset_hash


class TestDatasetHashing(TestCase):
    def setUp(self) -> None:
        self.orig = pd.DataFrame(
            [[1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6]], columns=["a", "b", "c"]
        )

    def test_shuffle_rows(self):
        shuf = self.orig.copy(deep=True)
        shuf = shuf.reindex(np.random.permutation(len(shuf)))

        self.assertEqual(
            dataset_hash(self.orig, 1), dataset_hash(shuf, 1), "Row order affects hash"
        )

    def test_shuffle_columns(self):
        shuf = self.orig.copy(deep=True)
        shuf = shuf[[c for c in shuf if c not in ["c"]] + ["c"]]

        self.assertEqual(
            dataset_hash(self.orig, 1), dataset_hash(shuf, 1), "Column order affects hash"
        )

    def test_single_cell_change(self):
        shuf = self.orig.copy(deep=True)
        shuf.iat[0, 2] = 6
        print(shuf)

        self.assertNotEqual(
            dataset_hash(self.orig, 1), dataset_hash(shuf, 1), "Row order affects hash"
        )



from typing import Callable, Dict, List, Tuple


class DatasetTransformation:
    """
    A Dataset Transformation object.

    :param func: Callable the function that modifies the Dataset.

    :param data_kwargs: Dict the data inputs to the function as keyword arguments. If they are outputs of a previous
        DatasetTransformation in the same series you should use the value "inherit". See the example for more details.

    :param kwargs: Dict any other keyword arguments to the function. Must be specified even if empty.

    :param outputs: List the name of the outputs. These should be the same as the input names of any
        DatasetTransformation that follows. See the example for more details.

    :param split: Optional[str] the split of the dataset. If the function does not split the dataset this may be ignored

    Example::
        >>> import pandas as pd
        >>>
        >>> def drop_correlated(data, thresh):
        >>>     import numpy as np
        >>>
        >>>     # calculate correlations
        >>>     corr_matrix = data.corr().abs()
        >>>     # get the upper part of correlation matrix
        >>>     upper = corr_matrix.where(np.triu(np.ones(corr_matrix.shape), k=1).astype(np.bool))
        >>>
        >>>     # columns with correlation above threshold
        >>>     redundant = [column for column in upper.columns if any(upper[column] >= thresh)]
        >>>     print(f"Columns to drop with correlation > {thresh}: {redundant}")
        >>>     data.drop(columns=redundant, inplace=True)
        >>>     return data
        >>>
        >>> def drop_nulls(df, threshold):
        >>>     cols = [x for x in df.columns if df[x].isnull().sum() / df.shape[0] > threshold]
        >>>     return df.drop(columns=cols)
        >>>
        >>> sample_df = pd.read_csv("some_data.csv")
        >>> corr_threshold = 0.97
        >>> sample_df1 = drop_correlated(data=sample_df, thresh=corr_threshold)
        >>>
        >>> null_threshold = 0.9
        >>> sample_df1 = drop_nulls(df=sample_df1, threshold=null_threshold)
        >>>
        >>> transformations = [
        >>>     DatasetTransformation(
        >>>         drop_correlated, {"data": sample_df}, {"thresh": corr_threshold}, ["df"]
        >>>     ),
        >>>     DatasetTransformation(
        >>>         drop_nulls, {"df": "inherit"}, {"threshold": null_threshold}, ["df"]
        >>>     ),


    """

    def __init__(
        self, func: Callable, data_kwargs: Dict, kwargs: Dict, outputs: List, split: str = None
    ):
        self.func = func
        self._data_kwargs = data_kwargs
        self.kwargs = kwargs
        self.outputs = outputs
        self.split = split

    def __call__(self, previous_output: Dict) -> Dict:
        # preprocess and data substitution.
        self._substitute_data_kwargs(previous_output)
        # runs the transformation function and returns the results (maybe only the named outputs (in self.outputs))
        outputs = self.func(**self._data_kwargs, **self.kwargs)
        outputs = self._process_outputs(outputs)
        return outputs

    def _process_outputs(self, outputs) -> Dict:
        # bind outputs
        results = dict()
        for key, val in zip(
            self.outputs,
            [outputs] if not (isinstance(outputs, Tuple) or isinstance(outputs, List)) else outputs,
        ):
            # strip out any outputs that have None for the key.
            if key is not None:
                results[key] = val
        return results

    def _substitute_data_kwargs(self, previous_output) -> None:
        for key, val in self._data_kwargs.items():
            if isinstance(val, str) and val == "inherit":
                self._data_kwargs[key] = previous_output[key]

    @property
    def data_kwargs(self) -> Dict:
        """Getter for data_kwargs that empties the data - to prevent uploading the data multiple times"""
        emptied_data_kwargs = dict()
        for key, val in self._data_kwargs.items():
            emptied_data_kwargs[key] = key
        return emptied_data_kwargs

    @property
    def raw_data_kwargs(self) -> Dict:
        """Returns the _data_kwargs without emptying them for the cases where that is needed."""
        return self._data_kwargs



from unittest import TestCase

from ..model_management import ModelEncodeError, ModelManagers, serialize


class Test(TestCase):
    def test_get_model_manager_framework_not_installed_exception(self):
        self.assertRaises(ModelEncodeError, serialize, "random model", ModelManagers.NOT_IMPORTED)

    def test_get_model_manager_sklearn_installed(self):
        self.assertIsNotNone(
            ModelManagers.SKLEARN.value,
            msg="Failed to instantiate framework",
        )

    def test_get_model_manager_xgboost_installed(self):
        self.assertIsNotNone(
            ModelManagers.XGBOOST.value,
            msg="Failed to instantiate framework",
        )

    def test_get_model_manager_lightgbm_installed(self):
        self.assertIsNotNone(
            ModelManagers.LIGHTGBM.value,
            msg="Failed to instantiate framework",
        )

    def test_get_model_manager_tensorflow_installed(self):
        self.assertIsNotNone(
            ModelManagers.TENSORFLOW.value,
            msg="Failed to instantiate framework",
        )



import os
import unittest
from unittest import TestCase, mock

from ..core.transmission import RequestWrapper

SERVER_ROOT = "http://mock_server.com"
PATH_EXISTING_FILE = "/existing_file"
PATH_MISSING_FILE = "/missing_file"
SAMPLE_FILE_CONTENT = b"some content"


def mocked_server_get_file(*args, **kwargs):
    class MockResponse:
        def __init__(self, content=SAMPLE_FILE_CONTENT, status_code=200, reason=None, text=None):
            self.status_code = status_code
            self.content = content
            self.reason = reason
            self.text = text

    if args[0] == f"{SERVER_ROOT}{PATH_EXISTING_FILE}":
        return MockResponse()
    if args[0] == f"{SERVER_ROOT}{PATH_MISSING_FILE}":
        return MockResponse(status_code=400, reason="File missing", text="HTML")


class TestTransmission(unittest.TestCase):
    def setUp(self) -> None:
        pass


class TestRequestWrapper(TestCase):
    def setUp(self) -> None:
        self.data_save_path = "./tmp"

        self.request_wrapper = RequestWrapper(SERVER_ROOT)

    def tearDown(self) -> None:
        try:
            os.remove(self.data_save_path)
        except FileNotFoundError:
            pass

    @mock.patch("requests.get", side_effect=mocked_server_get_file)
    def test_load_existing_file(self, mock_get):
        self.request_wrapper.load_file(f"{PATH_EXISTING_FILE}", self.data_save_path)
        with open(self.data_save_path, "rb") as f:
            self.assertEqual(
                f.read(),
                SAMPLE_FILE_CONTENT,
                msg="File not downloaded and saved to specified path",
            )

    @mock.patch("requests.get", side_effect=mocked_server_get_file)
    def test_load_missing_file(self, mock_get):
        self.assertRaises(
            Exception, self.request_wrapper.load_file, f"{PATH_MISSING_FILE}", self.data_save_path
        )


if __name__ == "__main__":
    unittest.main()



from seclea_ai.lib.seclea_utils.core.transmission import Transmission

from ..errors import throws_api_err

root = "/collection/projects"


@throws_api_err
def get_projects(transmission: Transmission) -> list:
    """Gets list of projects from the collection app given api_root

    @param transmission: Transmission type used for communication with server
    @return: List of projects or empty list
    """
    return transmission.get(f"{root}").json()



import logging
from io import BytesIO
from typing import Union

from typing.io import IO, BinaryIO

logger = logging.getLogger(__name__)

# Type Definitions #
BytesStream = Union[IO, IO[bytes], BinaryIO, BytesIO]

# Typing utility functions #
# TODO find a better way of recording imports that cant be imported.
_not_importable = set()


def get_type(obj) -> str:
    """
    Get type name from object to avoid dependencies. Credit wandb/client/get_model_manager.py for original code which has been modified.
    :param obj: The object needing to be typed
    :return: (str) String representation of the type.
    """
    name = f"{obj.__class__.__module__}.{obj.__class__.__name__}"
    if name in ["builtins.module", "__builtin__.module"]:
        return obj.__name__
    else:
        return name


def import_module(target, relative_to=None):
    """
    Imports a module.
    Credit wandb/client/get_model_manager.py
    :param target:
    :param relative_to:
    :return:
    """
    target_parts = target.split(".")
    target_depth = target_parts.count("")
    target_path = target_parts[target_depth:]
    target = target[target_depth:]
    fromlist = [target]
    if target_depth and relative_to:
        relative_parts = relative_to.split(".")
        relative_to = ".".join(relative_parts[: -(target_depth - 1) or None])
    if len(target_path) > 1:
        relative_to = ".".join(filter(None, [relative_to]) + target_path[:-1])
        fromlist = target_path[-1:]
        target = fromlist[0]
    elif not relative_to:
        fromlist = []
    mod = __import__(relative_to or target, globals(), locals(), fromlist)
    return getattr(mod, target, mod)


def get_module(name, required=None):
    """
    Return module or None. Absolute import is required.
    Credit wandb/client/get_model_manager.py with some minor modification.
    :param (str) name: Dot-separated module path. E.g., 'scipy.stats'.
    :param (str) required: A string to raise a ValueError if missing
    :return: (module|None) If import succeeds, the module will be returned.
    """
    if name not in _not_importable:
        try:
            return import_module(name)
        except Exception:
            _not_importable.add(name)
            msg = "Error importing optional module {}".format(name)
            if required:
                logger.exception(msg)
    if required is not None and name in _not_importable:
        raise ValueError(required)



import pickle  # nosec
from typing import Any, Dict

from .model_management import ModelManager


class SKLearnModelManager(ModelManager):
    @staticmethod
    def encode(model: Any) -> bytes:
        return pickle.dumps(model)  # nosec

    @staticmethod
    def decode(data: bytes) -> Any:
        """
        Loads a stored SKLearn model. As the model is stored with pickle and a certain version of SKLearn, there
        may be inconsistencies where different versions of SKLearn are used for pickling and unpickling.
        :param reference:
        :return:
        """
        return pickle.loads(data)  # nosec

    @staticmethod
    def get_params(model) -> Dict:
        """
        Extracts the parameters of the model.
        :param model: The model
        """

        return model.get_params()



from unittest import TestCase

import xgboost
from xgboost import Booster

from ..core.typing import get_module, get_type


class TestImports(TestCase):
    def test_get_module(self):
        xgb = get_module("xgboost")
        booster = xgb.Booster()
        self.assertTrue(isinstance(booster, Booster), "Should be xgboost Booster but is not.")

    def test_get_module_fail(self):
        django = get_module("adsfjaielkfd")
        self.assertIsNone(
            django, "adsfjaielkfd is imported but should be None. Check in clean environment."
        )

    def test_get_module_fail_required(self):
        with self.assertRaises(ValueError):
            _ = get_module("adsfjaielkfd", "Django required")

    def test_get_type(self):
        booster = xgboost.Booster()
        t = get_type(booster)
        self.assertEqual(t, "xgboost.core.Booster")



import os.path
import unittest

import pandas as pd

from ..dataset_management.dataset_utils import (
    get_dataset_hash,
    get_dataset_project_hash,
    get_project_hash,
)


class TestFileCompressor(unittest.TestCase):
    class Project:
        ...

    def setUp(self) -> None:
        base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        self.folder_path = os.path.join(base_dir, "tests")
        self.dataset = f"{self.folder_path}/example_files/filled_na_train.csv"
        self.df = pd.read_csv(self.dataset)
        self.series = self.df["DeviceType"]

    def test_get_dataset_project_hash(self):
        hash = get_dataset_project_hash(
            dataset=self.series,
            project=self.Project(),
        )
        self.assertIsInstance(hash, str)

    def test_get_dataset_hash(self):
        hash = get_dataset_hash(
            dataset=self.series,
        )
        self.assertIsInstance(hash, str)

    def test_get_project_hash(self):
        hash = get_project_hash(
            project=self.Project(),
        )
        self.assertIsInstance(hash, str)



class APIError(Exception):
    pass


def throws_api_err(f):
    def call(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            raise APIError(e)

    return call



import unittest

from ..core.transformations import decode_func, encode_func


class TestTransformations(unittest.TestCase):
    def test_encode_decode_function_no_args(self):
        def test_func(a):
            return 12 * 3 + a

        encoded = encode_func(test_func)
        decoded, args, kwargs = decode_func(encoded)
        self.assertEqual(decoded(2, *args, **kwargs), test_func(2))

    def test_encode_decode_function_args_only(self):
        def test_func(a, b):
            return a + b

        test_args = [1, -4]
        encoded = encode_func(test_func, args=test_args)
        decoded, args, kwargs = decode_func(encoded)
        self.assertEqual(decoded(*args, **kwargs), test_func(*test_args))

    def test_encode_decode_function_kwargs_only(self):
        def test_func(a, b):
            return a + (2 * b)

        test_kwargs = {"a": 1, "b": 3}

        encoded = encode_func(test_func, kwargs=test_kwargs)
        decoded, args, kwargs = decode_func(encoded)

        self.assertEqual(decoded(*args, **kwargs), test_func(**test_kwargs))  # add assertion here

    def test_encode_decode_function_args_kwargs(self):
        def test_func(a, b, c, d):
            return a - b + (c * d)

        test_args = [1, 4]
        test_kwargs = {"c": 3, "d": 5}

        encoded = encode_func(test_func, args=test_args, kwargs=test_kwargs)
        decoded, args, kwargs = decode_func(encoded)

        self.assertEqual(
            decoded(*args, **kwargs), test_func(*test_args, **test_kwargs)
        )  # add assertion here


if __name__ == "__main__":
    unittest.main()



import io
import os
from typing import Union

from typing.io import BinaryIO

from .compression import CompressionFactory


class ToVariableWriter:
    def __init__(self):
        self._data_bytes = []

    @property
    def data_bytes(self):
        return b"".join(self._data_bytes)

    @property
    def data_str(self):
        return self.data_bytes.decode("utf-8")

    def write(self, d: bytes):
        self._data_bytes.append(d)


def save_object(
    obj: Union[bytes, BinaryIO],
    file_name: str,
    path: str,
    compression: CompressionFactory = CompressionFactory.NONE,
) -> str:
    compression = compression.value()
    if isinstance(obj, bytes):
        obj = io.BytesIO(obj)
    file_name = f"{file_name}{compression.extension}"
    save_path = os.path.join(path, file_name)
    with open(save_path, "wb") as out_file:
        compression.compress(read_stream=obj, write_stream=out_file)
    return save_path


def load_object(path, as_bytes=True) -> Union[str, bytes]:
    compression_extension = path.split(".")[-1]
    compression = CompressionFactory[compression_extension].value()
    to_var = ToVariableWriter()
    with open(path, "rb") as in_file:
        compression.decompress(read_stream=in_file, write_stream=to_var)
        in_file.close()
    if as_bytes:
        return to_var.data_bytes
    return to_var.data_str



# import os
# import unittest
# from pathlib import Path
# from unittest import mock
# from unittest.mock import mock_open, patch
#
# import responses
# from  import RequestWrapper
#
# from seclea_ai.authentication import AuthenticationService
# from seclea_ai.exceptions import AuthenticationError
#
#
# class TestAuthenticationService(unittest.TestCase):
#     def setUp(self) -> None:
#         pass
#
#     @responses.activate
#     def test_refresh_token(self):
#         responses.add(
#             method=responses.POST,
#             url="http://localhost:8000/api/token/refresh/",
#             json={"access": "dummy_access_token"},
#             status=200,
#         )
#
#         auth_service = AuthenticationService(
#             transmission=RequestWrapper(server_root_url="http://localhost:8000")
#         )
#         with patch(
#             "builtins.open",
#             new=mock_open(read_data='{"refresh": "dummy_refresh_token"}\n'),
#         ) as mock_file:
#             creds = auth_service._refresh_token()
#         mock_file.assert_called()
#         self.assertEqual(
#             creds, {"Authorization": "Bearer dummy_access_token"}, msg="Auth doesn't match"
#         )
#
#     @responses.activate
#     def test_refresh_token_expired_token(self):
#         with self.assertRaises(AuthenticationError):
#             responses.add(
#                 method=responses.POST,
#                 url="http://localhost:8000/api/token/refresh/",
#                 body="Not Authorised",
#                 status=403,
#             )
#
#             auth_service = AuthenticationService(
#                 transmission=RequestWrapper(server_root_url="http://localhost:8000")
#             )
#             with patch(
#                 "builtins.open",
#                 new=mock_open(read_data='{"refresh": "dummy_refresh_token"}\n'),
#             ):
#                 auth_service._refresh_token()
#
#     @responses.activate
#     @mock.patch("seclea_ai.authentication.getpass", return_value="test_pass")
#     @mock.patch("builtins.input", autospec=True, return_value="test_user")
#     def test_login(self, mock_input, mock_getpass):
#         responses.add(
#             method=responses.POST,
#             url="http://localhost:8000/api/token/obtain/",
#             json={"access": "dummy_access_token", "refresh": "dummy_refresh_token"},
#             status=200,
#         )
#
#         auth_service = AuthenticationService(
#             transmission=RequestWrapper(server_root_url="http://localhost:8000")
#         )
#
#         with patch(
#             "builtins.open",
#             new=mock_open(read_data='{"refresh": "dummy_refresh_token"}\n'),
#         ) as mock_file:
#             creds = auth_service.login()
#         self.assertEqual(
#             creds, {"Authorization": "Bearer dummy_access_token"}, msg="Auth doesn't match"
#         )
#         mock_input.assert_called_once()
#         mock_getpass.assert_called_once()
#         mock_file.assert_called_with(os.path.join(Path.home(), ".seclea/config"), "w+")
#
#     @responses.activate
#     @mock.patch("seclea_ai.authentication.getpass", return_value="test_pass")
#     @mock.patch("builtins.input", autospec=True, return_value="test_user")
#     @mock.patch("seclea_ai.authentication.os.path")
#     @mock.patch("seclea_ai.authentication.os")
#     def test_handle_auth_refresh_expire(self, mock_os, mock_path, mock_input, mock_getpass):
#         responses.add(
#             method=responses.POST,
#             url="http://localhost:8000/api/token/obtain/",
#             json={"access": "dummy_access_token", "refresh": "dummy_refresh_token"},
#             status=200,
#         )
#         responses.add(
#             method=responses.POST,
#             url="http://localhost:8000/api/token/refresh/",
#             body="Not Authorised",
#             status=403,
#         )
#         mock_path.isfile.return_value = False
#         mock_os.mkdir.return_value = None
#
#         auth_service = AuthenticationService(
#             transmission=RequestWrapper(server_root_url="http://localhost:8000")
#         )
#
#         with patch(
#             "builtins.open",
#             new=mock_open(
#                 read_data='{"refresh": "dummy_refresh_token", "username": "test_user"}\n'
#             ),
#         ) as mock_file:
#             creds = auth_service.handle_auth()
#         self.assertEqual(
#             creds, {"Authorization": "Bearer dummy_access_token"}, msg="Auth doesn't match"
#         )
#         mock_path.isfile.assert_called_once()
#         mock_os.mkdir.assert_called()
#         mock_input.assert_called_once()
#         mock_getpass.assert_called_once()
#         mock_file.assert_called()



from abc import ABC, abstractmethod
from typing import Any


class ModelManager(ABC):
    @staticmethod
    @abstractmethod
    def encode(model: Any) -> bytes:
        """
        Saves a model at the specified reference.
        :param model: model to save
        :return: Where it was saved.
        """
        pass

    @staticmethod
    @abstractmethod
    def decode(data: bytes) -> Any:
        """
        :param data:
        :return: The model.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_params(model):
        """
        Gets the model parameters for a given model.
        """
        pass



import os.path
import unittest
import uuid

from ..core.compression import FileCompressor, Zstd

base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
folder_path = os.path.join(base_dir, "tests")


class TestZstd(unittest.TestCase):
    def setUp(self) -> None:
        self.comp = Zstd()

        self.original = f"{folder_path}/example_files/bee_movie.txt"

    def test_compress_decompress(self):
        decompressed_path = f"{folder_path}/example_files/bee_movie_decompressed.txt"
        compressed_path = f"{folder_path}/example_files/bee_movie_compressed.txt"
        orig_size = os.path.getsize(self.original)
        with open(self.original, "rb") as fs, open(compressed_path, "wb") as fo:
            self.comp.compress(fs, fo)
        self.assertLess(
            os.path.getsize(compressed_path), orig_size, "File post compression not smaller"
        )
        with open(compressed_path, "rb") as ch, open(decompressed_path, "wb") as co:
            self.comp.decompress(ch, co)
        self.assertEqual(
            os.path.getsize(decompressed_path),
            orig_size,
            "File post decompression not original size",
        )

        os.remove(compressed_path)
        os.remove(decompressed_path)


class TestFileCompressor(unittest.TestCase):
    def setUp(self) -> None:
        self.file_compressor = FileCompressor()
        self.original = f"{folder_path}/example_files/bee_movie.txt"

    def test_compression(self):
        compressed_path = f"{folder_path}/example_files/bee_movie_compressed{uuid.uuid4()}.txt"
        self.file_compressor.compression(file_path_in=self.original, file_path_out=compressed_path)
        self.assertLess(os.path.getsize(compressed_path), os.path.getsize(self.original))
        os.remove(compressed_path)

    def test_decompression(self):
        decompressed_path = f"{folder_path}/example_files/bee_movie_decompressed{uuid.uuid4()}.txt"
        compressed_path = f"{folder_path}/example_files/bee_movie_compressed{uuid.uuid4()}.txt"
        self.file_compressor.compression(file_path_in=self.original, file_path_out=compressed_path)
        self.file_compressor.decompression(
            file_path_in=compressed_path, file_path_out=decompressed_path
        )
        self.assertEqual(os.path.getsize(self.original), os.path.getsize(decompressed_path))
        self.assertLess(os.path.getsize(compressed_path), os.path.getsize(decompressed_path))
        os.remove(compressed_path)
        os.remove(decompressed_path)


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import Any

import uuid
import boto3
import dill as pickle  # nosec
import numpy as np
from PIL import Image
import pandas as pd

S3_BUCKET_NAME = "S3_BUCKET_NAME"


class DatasetWorker:
    PANDAS = "pandas"
    DILL = "dill"
    NUMPY = "numpy"
    FEATHER = "feather"
    PILLOW = "pillow"
    ZIPFILE = "zipfile"
    TARFILE = "tarfile"

    class Mode(Enum):
        S3 = 1
        LOCAL = 2

    mode = Mode.LOCAL

    def get_lib(self, file_type: str):
        FILE_OPENER = {
            # Single files
            self.PANDAS: ["csv", "txt", "parquet"],
            self.DILL: ["pickle"],
            self.NUMPY: ["npy"],
            self.FEATHER: ["feather"],
            self.PILLOW: ["jpg", "png"],
            # Archives
            self.ZIPFILE: ["gzip", "bzip", "7zip"],
            self.TARFILE: ["tar_gz", "tar_bz"],
            None: [  # We will work with those types in future
                # Single files
                "dhf5",
                "netcdf",
                "json",
                "orc",
                "avro",
                "pdf",
                "nii",
                "dicom",
            ],
        }

        for lib, file_types in FILE_OPENER.items():
            if file_type in file_types:
                return lib
        raise Exception(f"We didn't work with {file_type} file type")

    def open_pandas(self, file_path, file_type):
        if file_type == "parquet":
            return pd.read_parquet(file_path, engine="pyarrow")
        else:
            return pd.read_csv(file_path)

    def open_dill(self, file_path):
        with open(file_path, "rb") as f:
            # TODO [SECURITY] sign pickle package against authenticated users/organizations
            return pickle.load(f)  # nosec

    def open_numpy(self, file_path):
        return np.load(file_path)

    def open_feather(self, file_path):
        return pd.read_feather(file_path)

    def open_pillow(self, file_path):
        return Image.open(file_path)

    def load_dataset(self, file_path: str):
        file_type = file_path.split(".")[-1]
        lib = self.get_lib(file_type=file_type)
        return self.open_dataset(lib=lib, file_path=file_path)

    def save_on_s3(self, file_path):
        file_name = file_path.split("/")[-1]
        file_path_on_s3 = f"data/{file_name}"

        client = boto3.client("s3")
        client.put_object(
            Body=open(file_path, "rb"),
            Bucket=S3_BUCKET_NAME,
            Key=file_path_on_s3,
        )

        return file_path_on_s3

    def save_pandas(self, dataset: Any, file_path: str, file_type: str):
        if file_type == "parquet":
            dataset.to_parquet("dataset.parquet.gzip", compression="gzip")
        else:
            dataset.to_csv(file_path)

    def save_dill(self, dataset: Any, file_path: str):
        with open(file_path, "wb") as outp:
            pickle.dump(dataset, outp, pickle.HIGHEST_PROTOCOL)

    def save_numpy(self, dataset: Any, file_path: str):
        np.savetxt(file_path, dataset, delimiter=",")

    def save_feather(self, dataset: Any, file_path: str):
        dataFrame = pd.DataFrame(data=dataset)
        dataFrame.to_feather(file_path)

    def save_pillow(self, dataset: Any, file_path: str):
        dataset.save(file_path)

    def open_dataset(self, lib: str, file_path: str):
        file_type = file_path.split(".")[0]
        if lib == self.PANDAS:
            return self.open_pandas(file_path=file_path, file_type=file_type)
        elif lib == self.DILL:
            return self.open_dill(file_path=file_path)
        elif lib == self.NUMPY:
            return self.open_numpy(file_path=file_path)
        elif lib == self.FEATHER:
            return self.open_feather(file_path=file_path)
        elif lib == self.PILLOW:
            return self.open_pillow(file_path=file_path)

    def save_dataset(self, dataset: Any, file_type: str, path: str) -> str:
        lib = self.get_lib(file_type=file_type)
        file_path = f"{path}_{uuid.uuid4()}.{file_type}"

        if lib == self.PANDAS:
            self.save_pandas(dataset=dataset, file_path=file_path, file_type=file_type)
        elif lib == self.DILL:
            self.save_dill(dataset=dataset, file_path=file_path)
        elif lib == self.NUMPY:
            self.save_numpy(dataset=dataset, file_path=file_path)
        elif lib == self.FEATHER:
            self.save_feather(dataset=dataset, file_path=file_path)
        elif lib == self.PILLOW:
            self.save_pillow(dataset=dataset, file_path=file_path)

        if self.mode == self.Mode.S3:
            return self.save_on_s3(file_path=file_path)
        return file_path



from typing import Any, Dict

import xgboost as xgb

from .model_management import ModelManager
from ..core import CustomNamedTemporaryFile


class XGBoostModelManager(ModelManager):
    @staticmethod
    def encode(model: Any) -> bytes:
        # save to temp file then use manager to store.
        with CustomNamedTemporaryFile() as temp:
            # this could be from SKlearnAPI or LearningAPI which have significant differences
            model.save_model(temp.name)
            with open(temp.name, "rb") as read_temp:
                return read_temp.read()

    @staticmethod
    def decode(data: bytes) -> Any:
        """
        Loads a stored XGBoost model. Note this will always return a Booster (LearningAPI model) even if the original
        model was an SKLearn model. This will impact the methods available on the returned model.
        :param reference:
        :return: XGBoost.Booster model.
        """
        # need to know what kind of model -
        with CustomNamedTemporaryFile() as temp:
            temp.write(data)
            temp.flush()
            model = (
                xgb.Booster()
            )  # TODO need to be careful about customer usage - ie. do they use the best iteration for their model or not....
            model.load_model(temp.name)
        return model

    @staticmethod
    def get_params(model) -> Dict:
        """
        Extracts the parameters of the model.
        :param model: The model
        """

        return model.save_config()



class DecompressionError(Exception):
    def __init__(self, message):
        super(DecompressionError, self).__init__(message)


class CompressionError(Exception):
    def __init__(self, message):
        super(CompressionError, self).__init__(message)



import json
import os.path

from seclea_ai.lib.seclea_utils.core.transmission import Transmission

# from ..errors import throws_api_err
root = "/collection/datasets"


def test_json_valid(d):
    d = json.dumps(d)
    json.loads(d)
    pass


# @throws_api_err
def post_dataset(
    transmission: Transmission,
    dataset_file_path: str,
    project_pk: str,
    organization_pk: str,
    name: str,
    metadata: dict,
    dataset_pk: str,
    parent_dataset_hash: str = None,
):
    """
    @param transmission:
    @param dataset_file_path:
    @param project_pk:
    @param organization_pk:
    @param name:
    @param metadata:
    @param dataset_pk:
    @param parent_dataset_hash:
    @return:
    """
    dataset_queryparams = {"project": project_pk, "organization": organization_pk}
    test_json_valid(metadata)

    try:
        with open(dataset_file_path, "rb") as f:
            dataset_obj = {
                "project": (None, project_pk),
                "name": (None, name),
                "metadata": (None, json.dumps(metadata), "application/json"),
                "hash": (None, str(dataset_pk)),
                "parent": (None, parent_dataset_hash),
                "dataset": (os.path.basename(dataset_file_path), f),
            }
            res = transmission.send_file(
                url_path=f"{root}",
                obj=dataset_obj,
                query_params=dataset_queryparams,
            )
        return res
    except Exception as e:
        print(e)



from pathlib import Path

import pickledb as db  # nosec


class Storage:
    """Simple storage supporting pickledb for now."""

    def __init__(self, root: str = None, db_name="default.db"):
        """
        @param root: root path of db default is Path.home()/.seclea/
        @param db_name: name of database
        """
        if root is None:
            root = f"{Path.home()}/.seclea/"
        self._path = f"{root}/{db_name}"
        Path(root).mkdir(parents=True, exist_ok=True)
        # TODO this is unclosed during testing...
        self.db = db.load(self._path, True)

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path: str):
        """
        :param path:
        :return:
        """
        path_root = path.split("/")[-1].join("/")
        Path(path_root).mkdir(parents=True, exist_ok=True)
        self._path = path
        # TODO this is unclosed during testing...
        self.db = db.load(self._path, True)

    def write(self, key, val):
        self.db.set(key, val)

    def get(self, key):
        return self.db.get(key)



import base64
import types
from typing import Dict, List

import dill as pickle  # nosec


def encode_func(func, args: List = None, kwargs: Dict = None):
    if args is None:
        args = list()
    if kwargs is None:
        kwargs = dict()
    return base64.b64encode(
        pickle.dumps({"function": func.__code__, "args": args, "kwargs": kwargs})  # nosec
    ).decode("ascii")


def decode_func(func):
    args = None
    kwargs = None
    try:
        dic = pickle.loads(base64.b64decode(func))  # nosec
        function = dic["function"]
        f = types.FunctionType(function, globals(), "transformation1")
    except Exception as e:
        f = e
    else:
        try:
            args = dic["args"]
            kwargs = dic["kwargs"]
        except KeyError:
            raise ValueError("Function must contain either args or kwargs")
    return f, args, kwargs



import os
import unittest

import pandas as pd

from seclea_ai import SecleaAI


base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
folder_path = os.path.join(base_dir, "")
print(folder_path)


class TestSecleaAI(unittest.TestCase):

    # def test_check_features_different_names(self):
    #     parent_metadata = {
    #         "features": ["months_as_customer", "age", "policy_number", "policy_bind_date"]
    #     }
    #     test_dataset = pd.DataFrame([[3, 34, 339203, 34], [7, 32, 339103, 3]], columns=[1, 2, 3, 4])
    #     metadata = {}
    #     metadata = SecleaAI._check_features(dataset=test_dataset, metadata=metadata, parent_metadata=parent_metadata)
    #     print(metadata)
    #
    # def test_check_features_different_names_different_len(self):
    #     parent_metadata = {
    #         "features": ["months_as_customer", "age", "policy_number", "policy_bind_date"]
    #     }
    #     test_dataset = pd.DataFrame([[3, 34, 339203, 34], [7, 32, 339103, 3]], columns=[1, 2, 3, 4])
    #     metadata = {}
    #     metadata = SecleaAI._check_features(dataset=test_dataset, metadata=metadata, parent_metadata=parent_metadata)
    #     print(metadata)

    def test_get_dataset_type(self):
        # ARRANGE
        non_timeseries_dataset = pd.DataFrame(
            [[3, 34, 339203, 34], [7, 32, 339103, 3]], columns=[1, 2, 3, 4], index=[1, 2]
        )
        timeseries_dataset = pd.DataFrame(
            [[3, 34, 339203, 34], [7, 32, 339103, 3]],
            columns=[1, 2, 3, 4],
            index=["2022-01-01", "2022-01-02"],
        )

        # ACT
        res1 = SecleaAI._get_dataset_type(non_timeseries_dataset)
        res2 = SecleaAI._get_dataset_type(timeseries_dataset)

        # ASSERT
        self.assertEqual("tabular", res1)
        self.assertEqual("time_series", res2)


if __name__ == "__main__":
    unittest.main()



import os.path

from seclea_ai.lib.seclea_utils.core.transmission import Transmission

root = "/collection/model-states"


def post_model_state(
    transmission: Transmission,
    model_state_file_path: str,
    organization_pk: str,
    project_pk: str,
    training_run_pk: str,
    sequence_num: int,
    final_state,
):
    """

    @param transmission:
    @param model_state_file_path:
    @param organization_pk:
    @param project_pk:
    @param training_run_pk:
    @param sequence_num:
    @param final_state:
    @return:
    """
    with open(model_state_file_path, "rb") as f:
        res = transmission.send_file(
            url_path="/collection/model-states",
            obj={
                "project": (None, project_pk),
                "sequence_num": (None, sequence_num),
                "training_run": (None, training_run_pk),
                "final_state": (None, final_state),
                "state": (os.path.basename(model_state_file_path), f),
            },
            query_params={
                "organization": organization_pk,
                "project": project_pk,
            },
        )
    return res



import enum
import zlib
from abc import ABC, abstractmethod

from zstandard import ZstdCompressor, ZstdDecompressor

from .exceptions import CompressionError, DecompressionError
from .typing import BytesStream


class Compression(ABC):
    def __init__(self, chunk_size=256 * (10**3), compression_ext=".NONE"):
        self.chunk_size = chunk_size
        self.extension = compression_ext

    @abstractmethod
    def compress(self, read_stream: BytesStream, write_stream: BytesStream):
        write_stream.write(read_stream.read())

    @abstractmethod
    def decompress(self, read_stream: BytesStream, write_stream: BytesStream):
        write_stream.write(read_stream.read())


# Start of concrete implementations #
class Zstd(Compression):
    def __init__(self):
        super().__init__(compression_ext=".ZSTD")
        self.cctx = ZstdCompressor()
        self.dctx = ZstdDecompressor()

    def compress(self, read_stream, write_stream):
        try:
            read_stream.seek(0, 2)
            size = read_stream.tell()
            read_stream.seek(0, 0)
            self.cctx.copy_stream(read_stream, write_stream, size=size, write_size=self.chunk_size)
        except Exception:
            raise CompressionError("An error occurred during compression.")

    def decompress(self, read_stream, write_stream):
        try:
            with self.dctx.stream_reader(read_stream) as rs:
                rb = rs.read(self.chunk_size)
                while rb:
                    write_stream.write(rb)
                    rb = rs.read(self.chunk_size)
        except Exception as e:
            raise DecompressionError(f"An error occurred during decompression: {e}")


#   REFERENCE IF WE WANT TO HAVE MORE CONTROL OVER COMPRESSION IN THE FUTURE   ###
# while rb:
#     i += 1
#     print((i * self.chunk_size) / (2000 * (10 ** 6)))
#     t = time()
#     rb = read_stream.read(self.chunk_size)
#     print(rb)
#     print("read:", time() - t)
#     t = time()
#     out = self.chunker.compress(rb)
#     for o in out:
#         print(o)
#     print("compress:", time() - t)
#     t = time()
#
#     wr = self.cctx.stream_writer(write_stream, write_size=self.chunk_size)
#     print("decl:", time() - t)
#     t = time()
#     for o in out:
#         print("iter:", time() - t)
#         t = time()
#         wr.write(o)
#         print("write-chunk", time() - t)
#         t = time()
#
#     t = time()
#     print("write:", time() - t)
#
# for out in self.chunker.finish():
#     write_stream.write(out)


class CompressionFactory(enum.Enum):
    NONE = (Compression,)
    ZSTD = Zstd


class FileCompressor:
    def compression(self, file_path_in: str, file_path_out: str):
        with open(file_path_in, mode="rb") as fin, open(file_path_out, mode="wb") as fout:
            data = fin.read()
            compressed_data = zlib.compress(data, zlib.Z_BEST_COMPRESSION)
            fout.write(compressed_data)

    def decompression(self, file_path_in: str, file_path_out: str):
        with open(file_path_in, mode="rb") as fin, open(file_path_out, mode="wb") as fout:
            data = fin.read()
            decompressed_data = zlib.decompress(data)
            fout.write(decompressed_data)



import os
from unittest import TestCase

import lightgbm as lgb
import numpy as np
import pandas as pd

# import tensorflow as tf
import xgboost as xgb
from sklearn.ensemble import GradientBoostingClassifier

from ..model_management import ModelEncodeError, ModelManagers, deserialize, serialize

base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
folder_path = os.path.join(base_dir, "tests")


class TestSKLearnModelManager(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # read data
        data = pd.read_csv(f"{folder_path}/example_files/filled_na_train.csv").head(5)
        label = data["isFraud"].copy(deep=True)
        data = data.drop("isFraud", axis=1)
        # setup training params
        params = {
            "learning_rate": 0.2,
            "max_depth": 10,
            "min_samples_leaf": 24,
            "max_leaf_nodes": 496,
        }
        # train model
        cls._booster_skl = GradientBoostingClassifier(**params)
        cls._booster_skl.fit(data, label)
        cls.sample = data.iloc[[0]].to_numpy()

    def setUp(self) -> None:
        pass

    def test_serialize(self):
        serialize(self._booster_skl, ModelManagers.SKLEARN)
        self.assertRaises(ModelEncodeError, serialize, self._booster_skl, ModelManagers.XGBOOST)

    def test_deserialize(self):
        model_b = serialize(self._booster_skl, ModelManagers.SKLEARN)
        model = deserialize(model_b)
        self.assertTrue(
            np.allclose(
                model.predict_proba(self.sample), self._booster_skl.predict_proba(self.sample)
            )
        )


class TestXGBoostModelManager(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # read core
        data = pd.read_csv(f"{folder_path}/example_files/filled_na_train.csv").head(5)
        label = data["isFraud"].copy(deep=True)
        data = data.drop("isFraud", axis=1)
        cls.data = data
        # load to XGBoost format
        dtrain = xgb.DMatrix(data=data, label=label)
        # setup training params
        params = dict(max_depth=2, eta=1, objective="binary:logistic", nthread=4, eval_metric="auc")
        num_rounds = 5
        # train model
        cls._booster = xgb.train(params=params, dtrain=dtrain, num_boost_round=num_rounds)
        cls._booster_skl = xgb.XGBClassifier(objective="binary:logistic", use_label_encoder=False)
        cls._booster_skl.fit(data, label)
        cls.sample = data.iloc[[0]].to_numpy()

    def setUp(self) -> None:
        pass

    def test_save_learning_api(self):
        serialize(self._booster, ModelManagers.XGBOOST)

    def test_load_from_learning_api(self):
        model_str = serialize(self._booster, ModelManagers.XGBOOST)
        model = deserialize(model_str)

        # os.remove(new_file)
        # test that it executes as expected
        print(
            f"XGBoost from learning - Loaded: {model.inplace_predict(self.sample)} - Original: {self._booster.inplace_predict(self.sample)}"
        )
        self.assertTrue(
            np.allclose(
                model.inplace_predict(self.sample), self._booster.inplace_predict(self.sample)
            )
        )


class TestLightGBMModelManager(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # read data
        data = pd.read_csv(f"{folder_path}/example_files/filled_na_train.csv").head(5)
        label = data.pop("isFraud")
        # load to XGBoost format
        dtrain = lgb.Dataset(data=data, label=label, free_raw_data=True)
        # setup training params
        params = dict(max_depth=2, eta=1, objective="binary", num_threads=4, metric="auc")
        num_rounds = 5
        # train model
        cls._booster = lgb.train(params=params, train_set=dtrain, num_boost_round=num_rounds)
        cls._booster_skl = lgb.LGBMClassifier(objective="binary")
        cls._booster_skl.fit(data, label)
        cls.sample = data.iloc[[0]]

    def setUp(self) -> None:
        pass

    def test_save_learning_api(self):
        serialize(self._booster, ModelManagers.LIGHTGBM)

    def test_load_from_learning_api(self):
        model_b = serialize(self._booster, ModelManagers.LIGHTGBM)
        model = deserialize(model_b)
        self.assertTrue(np.allclose(model.predict(self.sample), self._booster.predict(self.sample)))
        # os.remove(temp)  # uncomment if not using FileManager


#     def test_save_sklearn_api(self):
#         temp = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
#         stored = self.manager.save_model(model=self._booster_skl, reference=temp)
#         self.assertTrue(os.path.exists(stored))
#         os.remove(stored)
#         # os.remove(temp)  # uncomment if not using FileManager
#
#     def test_load_from_sklearn_api(self):
#         temp = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
#         stored = self.manager.save_model(model=self._booster_skl, reference=temp)
#         model = self.manager.load_model(stored)
#         # os.remove(new_file)
#         # test that it executes as expected
#         print(
#             f"LGBM from SKL Loaded: {model.predict(self.sample)} - Original: {np.delete(self._booster_skl.predict_proba(self.sample), 0)}"
#         )
#         self.assertTrue(
#             np.allclose(
#                 model.predict(self.sample), np.delete(self._booster_skl.predict(self.sample), 0)
#             )
#         )
#         os.remove(stored)
#         # os.remove(temp)  # uncomment if not using FileManager


# class TestTensorflowModelManager(TestCase):
#     @staticmethod
#     def create_model(normalizer):
#
#         model = tf.keras.Sequential(
#             [
#                 normalizer,
#                 tf.keras.layers.Dense(512, activation="relu", input_shape=(111,)),
#                 tf.keras.layers.Dropout(0.2),
#                 tf.keras.layers.Dense(1),
#             ]
#         )
#
#         model.compile(
#             optimizer="adam",
#             loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
#         )
#
#         return model
#
#     @classmethod
#     def setUpClass(cls) -> None:
#         # read core
#         data = pd.read_csv(f"{folder_path}/example_files/filled_na_train.csv").head(5)
#         label = data["isFraud"].copy(deep=True)
#         data = data.drop("isFraud", axis=1)
#         cls.data = data
#         cls.label = label
#
#         normalizer = tf.keras.layers.Normalization(axis=-1)
#         normalizer.adapt(data)
#
#         # define model
#         tf_model = cls.create_model(normalizer=normalizer)
#         # train model
#         tf_model.fit(data, label, epochs=5)
#         cls._tf_model = tf_model
#
#         cls.sample = data.iloc[[0]].to_numpy()
#
#     def setUp(self) -> None:
#         pass
#
#     def test_save(self):
#         serialize(self._tf_model, ModelManagers.TENSORFLOW)
#
#     def test_load(self):
#         model_str = serialize(self._tf_model, ModelManagers.TENSORFLOW)
#         model = deserialize(model_str)
#
#         loaded = model.predict(self.sample)
#         orig = self._tf_model.predict(self.sample)
#
#         # test that it executes as expected
#         print(f"Tensorflow from learning - Loaded: {loaded} - Original: {orig}")
#         self.assertTrue(np.allclose(loaded, orig))



from .get_model_manager import deserialize
from ..core import load_object


def load_model_state(model_state_path: str):
    """
     # TODO: Convert part of this to a seclea utils?
     loads and deserializes model state into the model used in ML.
    :param model_state:
    :return:
    """
    model_b = load_object(model_state_path)
    return deserialize(model_b)



import traceback
from getpass import getpass

from seclea_ai.exceptions import AuthenticationError
from seclea_ai.lib.seclea_utils.core import Transmission
from .storage import Storage

try:
    import google.colab  # noqa F401

    IN_COLAB = True
except ImportError:
    IN_COLAB = False


class AuthenticationService:
    def __init__(self, transmission: Transmission):
        self._transmission = transmission
        self._db = Storage(db_name="auth_service", root="." if IN_COLAB else None)
        self._path_token_obtain = (
            "/api/token/obtain/"  # nosec - bandit thinks this is a pw or key..
        )
        self._path_token_refresh = (
            "/api/token/refresh/"  # nosec - bandit thinks this is a pw or key..
        )
        self._path_token_verify = (
            "/api/token/verify/"  # nosec - bandit thinks this is a pw or key..
        )
        self._key_token_access = "access_token"  # nosec - bandit thinks this is a pw or key..
        self._key_token_refresh = "refresh_token"  # nosec - bandit thinks this is a pw or key..

    def authenticate(self, transmission: Transmission = None, username=None, password=None):
        """
        Attempts to authenticate with server and then passes credential to specified transmission

        :param transmission: transmission service we wish to authenticate
        :return:
        """
        if not self.refresh_token():
            self._obtain_initial_tokens(username=username, password=password)
        if not self.verify_token():
            raise AuthenticationError("Failed to verify token")
        transmission.cookies = self._transmission.cookies

    def verify_token(self) -> bool:
        """
        Verifies if access token in database is valid
        :return: bool valid
        """
        self._transmission.cookies = {self._key_token_access: self._db.get(self._key_token_access)}
        try:
            response = self._transmission.send_json(url_path=self._path_token_verify, obj={})
        except TypeError:
            traceback.print_exc()
            return False
        return response.status_code == 200

    def refresh_token(self) -> bool:
        """
        Refreshes the access token by posting the refresh token.
        :return: bool success
        """
        if not self._db.get(self._key_token_refresh):
            return False
        self._transmission.cookies = {
            self._key_token_refresh: self._db.get(self._key_token_refresh)
        }
        response = self._transmission.send_json(url_path=self._path_token_refresh, obj={})
        self._save_response_tokens(response)
        return response.status_code == 200

    def _request_user_credentials(self) -> dict:
        """
        Gets user credentials manually
        :return:
        """
        return {"username": input("Username: "), "password": getpass("Password: ")}

    def _save_response_tokens(self, response) -> None:
        """
        Saves refresh and access tokens into db
        :param response:
        :return:
        """
        if self._key_token_refresh in response.cookies:
            self._db.write(self._key_token_refresh, response.cookies[self._key_token_refresh])
        if self._key_token_access in response.cookies:
            self._db.write(self._key_token_access, response.cookies[self._key_token_access])

    def _obtain_initial_tokens(self, username=None, password=None):
        if username is None or password is None:
            credentials = self._request_user_credentials()
        else:
            credentials = {"username": username, "password": password}
        response = self._transmission.send_json(url_path=self._path_token_obtain, obj=credentials)

        if response.status_code != 200:
            raise AuthenticationError(f"status:{response.status_code}, content:{response.content}")
        self._save_response_tokens(response)



import base64
import json
from enum import Enum

try:
    from .sklearn import SKLearnModelManager
except ValueError:
    SKLearnModelManager = None
try:
    from .lightgbm import LightGBMModelManager
except ValueError:
    LightGBMModelManager = None
except ModuleNotFoundError:
    LightGBMModelManager = None
try:
    from .xgboost import XGBoostModelManager
except ValueError:
    XGBoostModelManager = None
except ModuleNotFoundError:
    XGBoostModelManager = None
try:
    from .tensorflow import TensorflowModelManager
except ValueError:
    TensorflowModelManager = None
except ModuleNotFoundError:
    TensorflowModelManager = None
from typing import Any


class ModelManagers(Enum):
    XGBOOST = XGBoostModelManager
    LIGHTGBM = LightGBMModelManager
    SKLEARN = SKLearnModelManager
    TENSORFLOW = TensorflowModelManager
    NOT_IMPORTED = None


class ModelDecodeError(Exception):
    pass


class ModelEncodeError(Exception):
    pass


def serialize(model, model_manager: ModelManagers) -> bytes:
    """
    Serializes a model given a model manager so that it can be stored and reloaded later.
    @param model: The model to serialize
    @param model_manager: model manager to encode the model with
    @return: str of encoded model
    """
    try:
        if model_manager.value is None:
            raise ModelEncodeError(f"Framework {model_manager.name} is not imported")
        model_b = model_manager.value.encode(model)
        model_str = base64.b64encode(model_b).decode("ascii")
        return json.dumps([model_manager.name, model_str]).encode("utf-8")
    except Exception as e:
        raise ModelEncodeError(f"Issue encoding model: {e}")


def deserialize(data: bytes) -> Any:
    """
    returns a Loaded model given data (previously serialized)
    @param data: serialized model
    @return: loaded model
    """
    try:
        data_str = data.decode("utf-8")
        key, model_str = json.loads(data_str)
    except json.JSONDecodeError as e:
        raise ModelDecodeError("Failed to decode model:", e)
    except ValueError:
        raise ModelDecodeError(
            f"Wrong Encoding, expected list of two elements: key,val, got: {json.loads(data_str)}"
        )

    try:
        model_manager = ModelManagers[key].value
        model_b = base64.b64decode(model_str)

    except KeyError as e:
        raise ModelDecodeError(f"Failed to retrieve model manager with key: {key}, err:{e}")
    except AttributeError as e:
        raise ModelDecodeError(f"Framework {ModelManagers[key].name} is not imported, err:{e}")
    except Exception as e:
        raise ModelDecodeError(f"Unknown issue, potential decode implementation error:{e}")
    try:
        return model_manager.decode(model_b)
    except Exception as e:
        raise ModelDecodeError(
            f"Error decoding model, Framework:{model_manager.__class__}, err:{e}"
        )



import os
import uuid
from io import StringIO
from pathlib import Path

import pandas as pd
from pandas import DataFrame

from ..core import CompressionFactory, load_object, save_object


def save_dataset(dataset, file_name, path):
    """
    To be extended to support datasets other than csv.
    """
    Path(path).mkdir(parents=True, exist_ok=True)
    tmp_path = os.path.join(path, uuid.uuid4().__str__())
    dataset.to_csv(tmp_path, index=True)
    rb = open(tmp_path, "rb")
    save_path = save_object(rb, file_name=file_name, path=path, compression=CompressionFactory.ZSTD)
    rb.close()
    os.remove(tmp_path)
    return save_path


def load_dataset(dataset_path: str, index_col=None):
    """
     # TODO: Extend to different types of datasets
    :param model_dataset:
    :return:
    """

    dataset_str = load_object(dataset_path, as_bytes=False)
    dataset_stream = StringIO(dataset_str)
    df = pd.read_csv(dataset_stream, index_col=index_col)
    return df


def column_hash(dataset: DataFrame) -> int:
    """Invariant to ordering of columns/rows"""
    sum = 0
    for col in dataset.columns:
        sum += int(pd.util.hash_pandas_object(dataset[col]).sum())
    return sum


def dataset_hash(dataset, project: int) -> str:
    return str(hash(column_hash(dataset) + project))


def get_dataset_project_hash(dataset, project) -> str:
    return str(hash(bytes(dataset)) + hash(project))


def get_dataset_hash(dataset) -> str:
    return str(hash(bytes(dataset)))


def get_project_hash(project) -> str:
    return str(hash(project))



try:
    from setuptools import find_packages, setup
except ImportError:
    from distutils.core import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()
setup(
    name="seclea_ai",
    version="0.0.50",
    author="octavio",
    author_email="octavio.delser@gmail.com",
    description="Seclea integration.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seclea/seclea_ai",
    packages=find_packages(exclude=["**test", "**example_files"]),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "dill>=0.3.4",
        "pandas>=1.3.0",
        "pickleDB>=0.9.2",
        "requests>=2.0.0",
        "zstandard>=0.15.2",
    ],
)



import json
import os
from abc import ABC
from typing import Any, Dict
from urllib.parse import urlencode

import requests
from requests import Response


# Interface declaration #
class Transmission(ABC):
    """
    Manages communication with a server, included cookies, cache, headers, paths etc...
    """

    def __init__(self, server_root_url):
        self._server_root = server_root_url
        self._headers = {}
        self._cookies = {}

    @property
    def cookies(self) -> dict:
        return self._cookies.copy()

    @cookies.setter
    def cookies(self, new_cookies: Dict) -> None:
        self._cookies.update(new_cookies)
        self.headers = {"HTTP_COOKIE": json.dumps(self.cookies)}

    @property
    def headers(self) -> Dict:
        """
        Returns a copy of the headers.
        :return: Dict A copy of the headers
        """
        return self._headers.copy()

    @headers.setter
    def headers(self, new_headers: Dict) -> None:
        """
        Adds new headers to the classes stored headers.
        :param new_headers: dict The new headers to add
        :return: None
        """
        self._headers = {**self._headers, **new_headers}

    def load_file(self, url_path: str, file_path: str, query_params: Dict = None) -> None:
        """
        Loads a file from an endpoint to a specified file path.
        """
        pass

    def send_file_old(
        self, url_path: str, file_path: str, query_params: Dict = None, delete_file=False
    ) -> Any:
        """
        Sends a file from a specified path to an endpoint.
        """
        pass

    def get(self, url_path: str, query_params: Dict = None) -> Any:
        """
        Sends a get request to a specified endpoint.
        """
        pass

    def send_json(self, url_path: str, obj: Dict, query_params: Dict = None) -> Any:
        """
        Sends a json object to a specified endpoint.
        """
        pass

    def send_file(self, url_path: str, obj: Dict, query_params: Dict = None) -> Any:
        """
        Sends json with files
        files must be open in the obj.
        """
        pass


# Start of concrete implementations #
class RequestWrapper(Transmission):
    def __init__(self, server_root_url):
        super(RequestWrapper, self).__init__(server_root_url)

    def load_file(self, url_path: str, file_path: str, query_params: Dict = None) -> None:
        res = requests.get(self._server_root + url_path, headers=self.headers, params=query_params)
        if res.status_code == 200:
            with open(file_path, "wb") as f:
                f.write(res.content)
        else:
            raise Exception(
                f"Issue requesting file, status: {res.status_code}, Reason: {res.reason} {res.text}"
            )

    def send_file_old(
        self, url_path: str, file_path: str, query_params: Dict = None, delete_file=False
    ) -> Response:
        with open(file_path, "rb") as f:
            headers = self.headers
            headers.update(
                {
                    "format": "multipart",
                    "Content-Type": "multipart/json",
                    "HTTP_CONTENT_DISPOSITION": "attachment; filename=",
                }
            )
            request_path = f"{self._server_root}{url_path}"
            res = requests.post(
                request_path,
                files={"file": f},
                headers=headers,
                cookies=self.cookies,
                params=query_params,
            )
        if delete_file:
            os.remove(file_path)
        return res

    def patch(self, url_path: str, obj: Dict, query_params: Dict = None) -> Response:
        headers = self.headers
        headers.update(
            {
                "format": "json",
                "Content-Type": "application/json",
            }
        )
        request_path = f"{self._server_root}{url_path}"
        return requests.patch(
            request_path,
            data=obj,
            headers=headers,
            cookies=self.cookies,
            params=query_params,
        )

    def get(self, url_path: str, query_params: Dict = None) -> Response:
        request_path = f"{self._server_root}{url_path}"
        return requests.get(
            request_path, headers=self.headers, params=query_params, cookies=self.cookies
        )

    def send_file(self, url_path: str, obj: Dict, query_params: Dict = None) -> Response:
        request_path = f"{self._server_root}{url_path}?{urlencode(query_params)}"
        return requests.post(request_path, cookies=self.cookies, files=obj)

    def send_json(self, url_path: str, obj: Dict, query_params: Dict = None) -> Response:
        headers = self.headers
        headers["content-type"] = "application/json"
        request_path = f"{self._server_root}{url_path}"
        return requests.post(
            request_path,
            data=json.dumps(obj),
            headers=headers,
            cookies=self.cookies,
            params=query_params,
        )



import json
import os
import pathlib
import shutil
import tempfile
from typing import Any, Dict

import tensorflow as tf

from .model_management import ModelManager
from ..core import CustomNamedTemporaryFile, encode_dir_to_file, decode_file_to_dir


def clean_up(before: set, after: set):
    new = after.difference(before)
    for f in new:
        if pathlib.Path(f).is_file():
            os.remove(f)
        elif pathlib.Path(f).is_dir():
            shutil.rmtree(f)
        else:
            raise ValueError("Not file or directory - cannot clean up.")


# TODO see if we can improve the flow - will require rework of the flow for all models I think
class TensorflowModelManager(ModelManager):
    @staticmethod
    def encode(model: tf.keras.Model) -> bytes:
        # save to folder - SavedModel standard
        before = set(os.listdir())
        with tempfile.TemporaryDirectory() as temp_dir:
            tf.keras.models.save_model(model, temp_dir, save_format="tf")

            # we assume tar is available everywhere - it should be in almost all cases.
            with CustomNamedTemporaryFile() as temp:
                try:
                    with open(temp.name, "wb+") as f:
                        encode_dir_to_file(pathlib.PurePath(temp_dir), f)
                # TODO remove this if tempfile cleans up in error cases...
                except Exception:
                    clean_up(before=before, after=set(os.listdir()))
                    raise

                with open(temp.name, "rb") as read_temp:
                    return read_temp.read()

    @staticmethod
    def decode(data: bytes) -> Any:
        """
        Loads a stored XGBoost model. Note this will always return a Booster (LearningAPI model) even if the original
        model was an SKLearn model. This will impact the methods available on the returned model.
        :param data:
        :return: XGBoost.Booster model.
        """
        before = set(os.listdir())
        with CustomNamedTemporaryFile() as temp:
            with open(temp.name, "wb+") as write_temp:
                write_temp.write(data)

            with open(temp.name, "rb") as read_temp:
                dirname = decode_file_to_dir(read_temp)

            # load model
            try:
                model = tf.keras.models.load_model(dirname)
            except Exception:
                clean_up(before=before, after=set(os.listdir()))
                raise
            clean_up(before=before, after=set(os.listdir()))
            return model

    @staticmethod
    def get_params(model) -> Dict:  # TODO check if this actually gets the parameter
        """
        Extracts the parameters of the model.
        :param model: The model
        """
        return json.loads(model.to_json())



# Original Credit https://stackoverflow.com/questions/23212435/permission-denied-to-write-to-my-temporary-file/63173312#63173312
# Extended by Roger Milroy

import os
import tempfile
from pathlib import PurePath


class CustomNamedTemporaryFile:
    """
    This custom implementation is needed because of the following limitation of tempfile.NamedTemporaryFile:

    > Whether the name can be used to open the file a second time, while the named temporary file is still open,
    > varies across platforms (it can be so used on Unix; it cannot on Windows NT or later).
    """

    def __init__(self, mode="w+b", delete=True):
        self._mode = mode
        self._delete = delete
        self.name = None

    def __enter__(self):
        # Generate a random temporary file name
        self.name = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
        # Ensure the file is created
        open(self.name, "x").close()
        # Open the file in the given mode
        self._tempFile = open(self.name, self._mode)
        return self._tempFile

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._tempFile.close()
        if self._delete:
            os.remove(self.name)


FILE_START = b"-*-start-file-*-\n"
DIR_START = b"-*-start-dir-*-\n"
DIR_END = b"-*-end-dir-*-\n"


def encode_dir_to_file(dir_or_file: PurePath, file):
    """
    Encode a whole directory structure into a single structured file - a la tar.

    Recursively encode and return a useable/decodeable structure for us to turn back into a directory.

    NOTE - this is not designed for large files - it will overflow memory almost certainly.
    Will need a streaming version if we stick with this for any length of time.

    Also designed for use with decode_file_to_dir on the same system only. Not cross
    :param dir_or_file:t
    :param file: file handle to write to. opened with wb(+) as the mode.
    :return:
    """

    # encode dir
    if os.path.isdir(dir_or_file):
        # write start delim and filename
        file.write(DIR_START)
        file.write(dir_or_file.parts[-1].encode("utf-8") + b"\n")
        # encode each file recursively
        for name in os.listdir(dir_or_file):
            encode_dir_to_file(dir_or_file / name, file)
        # write end delim
        file.write(DIR_END)

    # encode file
    elif os.path.isfile(dir_or_file):
        # write start delim and filename
        file.write(FILE_START)
        file.write(dir_or_file.parts[-1].encode("utf-8") + b"\n")
        file.write(str(os.path.getsize(dir_or_file)).encode("utf-8") + b"\n")
        # write file
        with open(dir_or_file, "rb") as f:
            file.write(f.read())
    else:
        raise ValueError("Inputs can only be files or directories")


def decode_file_to_dir(file_handle) -> str:
    """
    Decode file to directory - placed in current directory unless otherwise specified.
    :param file_handle:
    :return:
    """

    # enter dir block create and enter directory
    # read line and pass to write func

    root = None

    while True:
        line = file_handle.readline()
        if line == b"":
            break
        elif line == FILE_START:
            line = file_handle.readline()
            file_name = line.decode("utf-8").rstrip("\n")
            line = file_handle.readline()
            size = int(line.decode("utf-8").rstrip("\n"))
            with open(file_name, "wb+") as write_f:
                write_f.write(file_handle.read(size))
        elif line == DIR_START:
            line = file_handle.readline()
            dir_name = line.decode("utf-8").rstrip("\n")
            os.mkdir(dir_name)
            if root is None:
                root = dir_name
            os.chdir(dir_name)
        elif line == DIR_END:
            os.chdir("..")  # todo make work on windows...
        else:
            raise ValueError("Something went wrong in decoding")

    if root is None:
        raise ValueError("Root name not defined - something went wrong encoding.")
    return root



import copy
from typing import Any, Dict

import lightgbm as lgb

from .model_management import ModelManager


class LightGBMModelManager(ModelManager):
    @staticmethod
    def encode(model: Any) -> bytes:
        # saves either the best or all iterations.. TODO think
        if isinstance(model, lgb.Booster):
            model_data = model.model_to_string().encode("utf-8")
        else:
            model_data = model.booster_.model_to_string().encode("utf-8")
        return model_data

    @staticmethod
    def decode(data: bytes) -> Any:
        """
        Loads a stored LightGBM model. Note this will always return a LightGBM Booster, even if the original model was
        an SKLearn model. This will impact the methods available on the returned model.
        :param data:
        :return: LightGBM.Booster model
        """
        model_str = data.decode()
        model = lgb.Booster(model_str=model_str)
        return model

    @staticmethod
    def get_params(model) -> Dict:
        """
        Extracts the parameters of the model.
        :param model: The model
        """

        return copy.deepcopy(model.params)



"""
Description for seclea_ai.py
"""
import copy
import inspect
import os
from pathlib import Path
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from pandas.errors import ParserError
from requests import Response

from seclea_ai.authentication import AuthenticationService
from seclea_ai.exceptions import AuthenticationError
from seclea_ai.lib.seclea_utils.core import (
    CompressionFactory,
    RequestWrapper,
    decode_func,
    encode_func,
    save_object,
)
import uuid
from .lib.seclea_utils.dataset_management.dataset_utils import save_dataset, dataset_hash
from seclea_ai.lib.seclea_utils.model_management.get_model_manager import ModelManagers, serialize
from seclea_ai.transformations import DatasetTransformation
from .svc.api.collection.dataset import post_dataset
from .svc.api.collection.model_state import post_model_state


def handle_response(res: Response, expected: int, msg: str) -> Response:
    """
    Handle responses from the server

    :param res: Response The response from the server.

    :param expected: int The expected HTTP status code (ie. 200, 201 etc.)

    :param msg: str The message to include in the Exception that is raised if the response doesn't have the expected
        status code

    :return: Response

    :raises: ValueError - if the response code doesn't match the expected code.

    """
    if not res.status_code == expected:
        raise ValueError(f"Response Status code {res.status_code}, expected:{expected}. \n{msg}")
    return res


class SecleaAI:
    def __init__(
        self,
        project_name: str,
        organization: str,
        platform_url: str = "https://platform.seclea.com",
        auth_url: str = "https://auth.seclea.com",
        username: str = None,
        password: str = None,
    ):
        """
        Create a SecleaAI object to manage a session. Requires a project name and framework.

        :param project_name: The name of the project

        :param platform_url: The url of the platform server. Default: "https://platform.seclea.com"

        :param auth_url: The url of the auth server. Default: "https://auth.seclea.com"

        :param username: seclea username

        :param password: seclea password

        :return: SecleaAI object

        :raises: ValueError - if the framework is not supported.

        Example::

            >>> seclea = SecleaAI(project_name="Test Project")
        """
        self._auth_service = AuthenticationService(RequestWrapper(auth_url))
        self._transmission = RequestWrapper(server_root_url=platform_url)
        self._auth_service.authenticate(self._transmission, username=username, password=password)
        self._project = None
        self._project_name = project_name
        self._organization = organization
        self._available_frameworks = {"sklearn", "xgboost", "lightgbm"}
        self._training_run = None
        self._cache_dir = os.path.join(Path.home(), f".seclea/{self._project_name}")
        self._init_project(project_name=project_name)
        print("success")

    def login(self, username=None, password=None) -> None:
        """
        Override login, this also overwrites the stored credentials in ~/.seclea/config.
        Note. In some circumstances the password will be echoed to stdin. This is not a problem in Jupyter Notebooks
        but may appear in scripting usage.

        :return: None

        Example::

            >>> seclea = SecleaAI(project_name="Test Project")
            >>> seclea.login()
        """
        success = False
        for i in range(3):
            try:
                self._auth_service.authenticate(
                    self._transmission, username=username, password=password
                )
                success = True
                break
            except AuthenticationError as e:
                print(e)
        if not success:
            raise AuthenticationError("Failed to login.")

    def upload_dataset_split(
        self,
        X: Union[DataFrame, np.ndarray],
        y: Union[DataFrame, np.ndarray],
        dataset_name: str,
        metadata: Dict,
        transformations: List[DatasetTransformation] = None,
    ) -> None:
        """
        Uploads a dataset.

        :param X: DataFrame The samples of the Dataset.

        :param y: Dataframe The labels of the Dataset

        :param dataset_name: The name of the Dataset

        :param metadata: Any metadata about the Dataset. Required keys are:
            "index" and "continuous_features"

        :param transformations: A list of DatasetTransformation's.

                        If your Dataset is large try and call this function more often with less DatasetTransformations
                        as the function currently requires (no. DatasetTransformations x Dataset size) memory.

                        See DatasetTransformation for more details.

        :return: None
        """
        dataset = self._assemble_dataset({"X": X, "y": y})
        # potentially fragile vvv TODO check this vvv
        metadata["outcome_name"] = y.name
        # try and extract the index automatically
        try:
            metadata["index"] = X.index.name
        except AttributeError:
            metadata["index"] = 0
        self.upload_dataset(dataset, dataset_name, metadata, transformations)

    def upload_dataset(
        self,
        dataset: Union[str, List[str], DataFrame],
        dataset_name: str,
        metadata: Dict,
        transformations: List[DatasetTransformation] = None,
    ) -> None:
        """
        Uploads a dataset.

        :param dataset: DataFrame, Path or list of paths to the dataset.
            If a list then they must be split by row only and all
            files must contain column names as a header line.

        :param dataset_name: The name of the dataset.

        :param metadata: Any metadata about the dataset. Note that if using a Path or list of Paths then if there is an
            index that you use when loading the data, it must be specified in the metadata.

        :param transformations: A list of DatasetTransformation's.

                        If your Dataset is large try call this function more often with less DatasetTransformations
                        as the function currently requires no. DatasetTransformations x Dataset size memory to function.

                        See DatasetTransformation for more details.

        :return: None

        Example:: TODO update docs
            >>> seclea = SecleaAI(project_name="Test Project")
            >>> dataset = pd.read_csv("/test_folder/dataset_file.csv")
            >>> dataset_metadata = {"index": "TransactionID", "outcome_name": "isFraud", "continuous_features": ["TransactionDT", "TransactionAmt"]}
            >>> seclea.upload_dataset(dataset=dataset, dataset_name="Multifile Dataset", metadata=dataset_metadata)

        Example with file::

            >>> seclea.upload_dataset(dataset="/test_folder/dataset_file.csv", dataset_name="Test Dataset", metadata={})
            >>> seclea = SecleaAI(project_name="Test Project", organization="Test Organization")

        Assuming the files are all in the /test_folder/dataset directory.
        Example with multiple files::

            >>> files = os.listdir("/test_folder/dataset")
            >>> seclea = SecleaAI(project_name="Test Project")
            >>> dataset_metadata = {"index": "TransactionID", "outcome_name": "isFraud", "continuous_features": ["TransactionDT", "TransactionAmt"]}
            >>> seclea.upload_dataset(dataset=files, dataset_name="Multifile Dataset", metadata=dataset_metadata)


        """
        # processing the final dataset - make sure it's a DataFrame
        if self._project is None:
            raise Exception("You need to create a project before uploading a dataset")

        if isinstance(dataset, List):
            dataset = self._aggregate_dataset(dataset, index=metadata["index"])
        elif isinstance(dataset, str):
            dataset = pd.read_csv(dataset, index_col=metadata["index"])

        dset_pk = dataset_hash(dataset, self._project)

        if transformations is not None:

            parent = self._assemble_dataset(transformations[0].raw_data_kwargs)

            #####
            # Validate parent exists and get metadata - can factor out
            #####
            parent_dset_pk = dataset_hash(parent, self._project)
            # check parent exists - throw an error if not.
            res = self._transmission.get(
                url_path=f"/collection/datasets/{parent_dset_pk}",
                query_params={"project": self._project, "organization": self._organization},
            )
            if not res.status_code == 200:
                raise AssertionError(
                    "Parent Dataset does not exist on the Platform. Please check your arguments and "
                    "that you have uploaded the parent dataset already"
                )
            parent_metadata = res.json()["metadata"]
            #####

            upload_queue = self._generate_intermediate_datasets(
                transformations=transformations,
                dataset_name=dataset_name,
                dset_pk=dset_pk,
                user_metadata=metadata,
                parent=parent,
                parent_metadata=parent_metadata,
            )

            # check for duplicates
            for up_kwargs in upload_queue:
                if dataset_hash(up_kwargs["dataset"], self._project) == up_kwargs["parent_hash"]:
                    raise AssertionError(
                        f"""The transformation {up_kwargs['transformation'].func.__name__} does not change the dataset.
                        Please remove it and try again."""
                    )
            # upload all the datasets and transformations.
            for up_kwargs in upload_queue:
                self._upload_dataset(**up_kwargs)  # TODO change
            return

        # this only happens if this has no transformations ie. it is a Raw Dataset.
        metadata_defaults_spec = dict(
            continuous_features=[],
            outcome_name=None,
            num_samples=len(dataset),
            favourable_outcome=None,
            unfavourable_outcome=None,
            dataset_type=self._get_dataset_type(dataset),
        )
        try:
            features = (
                dataset.columns
            )  # TODO - drop the outcome name but requires changes on frontend.
        except KeyError:
            # this means outcome was set to None
            features = dataset.columns

        metadata = self._ensure_required_metadata(
            metadata=metadata, defaults_spec=metadata_defaults_spec
        )
        automatic_metadata = dict(
            index=0 if dataset.index.name is None else dataset.index.name,
            split=None,
            features=list(features),
            categorical_features=list(
                set(list(features))
                - set(metadata["continuous_features"]).intersection(set(list(features)))
            ),
        )
        metadata = self._add_required_metadata(metadata=metadata, required_spec=automatic_metadata)

        metadata["categorical_values"] = [
            {col: dataset[col].unique().tolist()} for col in metadata["categorical_features"]
        ]

        self._upload_dataset(
            dataset=dataset,
            dataset_name=dataset_name,
            metadata=metadata,
            parent_hash=None,
            transformation=None,
        )

    def _generate_intermediate_datasets(
        self, transformations, dataset_name, dset_pk, user_metadata, parent, parent_metadata
    ):
        # setup for generating datasets.
        last = len(transformations) - 1
        upload_queue = list()
        parent_mdata = parent_metadata
        parent_dset = parent

        output = dict()
        # iterate over transformations, assembling intermediate datasets
        # TODO address memory issue of keeping all datasets
        for idx, trans in enumerate(transformations):
            output = trans(output)

            # construct the generated dataset from outputs
            dset = self._assemble_dataset(output)

            dset_metadata = copy.deepcopy(user_metadata)
            # validate and ensure required metadata
            metadata_defaults_spec = dict(
                continuous_features=parent_mdata["continuous_features"],
                outcome_name=parent_mdata["outcome_name"],
                num_samples=len(dset),
                favourable_outcome=parent_mdata["favourable_outcome"],
                unfavourable_outcome=parent_mdata["unfavourable_outcome"],
                dataset_type=self._get_dataset_type(dset),
            )
            dset_metadata = self._ensure_required_metadata(
                metadata=dset_metadata, defaults_spec=metadata_defaults_spec
            )
            try:
                features = (
                    dset.columns
                )  # TODO - drop the outcome name but requires changes on frontend.
            except KeyError:
                # this means outcome was set to None
                features = dset.columns

            automatic_metadata = dict(
                index=0 if dset.index.name is None else dset.index.name,
                split=trans.split if trans is not None else parent_mdata["split"],
                features=list(features),
                categorical_features=list(
                    set(list(features))
                    - set(dset_metadata["continuous_features"]).intersection(set(list(features)))
                ),
            )

            dset_metadata = self._add_required_metadata(
                metadata=dset_metadata, required_spec=automatic_metadata
            )

            dset_metadata["categorical_values"] = [
                {col: dset[col].unique().tolist()} for col in dset_metadata["categorical_features"]
            ]

            # constraints
            if not set(dset_metadata["continuous_features"]).issubset(
                set(dset_metadata["features"])
            ):
                raise ValueError(
                    "Continuous features must be a subset of features. Please check and try again."
                )

            dset_name = f"{dataset_name}-{trans.func.__name__}"  # TODO improve this.

            # handle the final dataset - check generated = passed in.
            if idx == last:
                if (
                    dataset_hash(dset, self._project) != dset_pk
                ):  # TODO create or find better exception
                    raise AssertionError(
                        """Generated Dataset does not match the Dataset passed in.
                                     Please check your DatasetTransformation definitions and try again.
                                     Try using less DatasetTransformations if you are having persistent problems"""
                    )
                else:
                    dset_name = dataset_name

            # add data to queue to upload later after final dataset checked
            upload_kwargs = {
                "dataset": copy.deepcopy(dset),  # TODO change keys
                "dataset_name": copy.deepcopy(dset_name),
                "metadata": dset_metadata,
                "parent_hash": dataset_hash(parent_dset, self._project),
                "transformation": copy.deepcopy(trans),
            }
            # update the parent dataset - these chained transformations only make sense if they are pushing the
            # same dataset through multiple transformations.
            parent_dset = copy.deepcopy(dset)
            parent_mdata = copy.deepcopy(dset_metadata)
            upload_queue.append(upload_kwargs)

        return upload_queue

    def upload_training_run_split(
        self,
        model,
        X_train: DataFrame,
        y_train: Union[DataFrame, Series],
        X_test: DataFrame = None,
        y_test: Union[DataFrame, Series] = None,
        X_val: Union[DataFrame, Series] = None,
        y_val: Union[DataFrame, Series] = None,
    ) -> None:
        """
        Takes a model and extracts the necessary data for uploading the training run.

        :param model: An ML Model instance. This should be one of {sklearn.Estimator, xgboost.Booster, lgbm.Boster}.

        :param X_train: Samples of the dataset that the model is trained on

        :param y_train: Labels of the dataset that the model is trained on.

        :param X_test: Samples of the dataset that the model is trained on

        :param y_test: Labels of the dataset that the model is trained on.

        :param X_val: Samples of the dataset that the model is trained on

        :param y_val: Labels of the dataset that the model is trained on.

        :return: None
        """
        train_dataset = self._assemble_dataset({"X": X_train, "y": y_train})
        test_dataset = None
        val_dataset = None
        if X_test is not None and y_test is not None:
            test_dataset = self._assemble_dataset({"X": X_test, "y": y_test})
        if X_val is not None and y_val is not None:
            val_dataset = self._assemble_dataset({"X": X_val, "y": y_val})
        self.upload_training_run(
            model=model,
            train_dataset=train_dataset,
            test_dataset=test_dataset,
            val_dataset=val_dataset,
        )

    def upload_training_run(
        self,
        model,
        train_dataset: DataFrame,
        test_dataset: DataFrame = None,
        val_dataset: DataFrame = None,
    ) -> None:
        """
        Takes a model and extracts the necessary data for uploading the training run.

        :param model: An ML Model instance. This should be one of {sklearn.Estimator, xgboost.Booster, lgbm.Boster}.

        :param train_dataset: DataFrame The Dataset that the model is trained on.

        :param test_dataset: DataFrame The Dataset that the model is trained on.

        :param val_dataset: DataFrame The Dataset that the model is trained on.

        :return: None

        Example::

            >>> seclea = SecleaAI(project_name="Test Project")
            >>> dataset = pd.read_csv(<dataset_name>)
            >>> model = LogisticRegressionClassifier()
            >>> model.fit(X, y)
            >>> seclea.upload_training_run(
                    model,
                    framework=seclea_ai.Frameworks.SKLEARN,
                    dataset_name="Test Dataset",
                )
        """
        self._auth_service.authenticate(self._transmission)

        # validate the splits? maybe later when we have proper Dataset class to manage these things.
        dataset_pks = [
            dataset_hash(dataset, self._project)
            for dataset in [train_dataset, test_dataset, val_dataset]
            if dataset is not None
        ]

        model_name = model.__class__.__name__

        framework = self._get_framework(model)

        # check the model exists upload if not
        model_type_pk = self._set_model(model_name=model_name, framework=framework)

        # check the latest training run
        training_runs_res = self._transmission.get(
            "/collection/training-runs",
            query_params={
                "project": self._project,
                "model": model_type_pk,
                "organization": self._organization,
            },
        )
        training_runs = training_runs_res.json()

        # Create the training run name
        largest = -1
        for training_run in training_runs:
            num = int(training_run["name"].split(" ")[2])
            if num > largest:
                largest = num
        training_run_name = f"Training Run {largest + 1}"

        # extract params from the model
        params = framework.value.get_params(model)

        # upload training run
        tr_res = self._upload_training_run(
            training_run_name=training_run_name,
            model_pk=model_type_pk,
            dataset_pks=dataset_pks,
            params=params,
        )
        # if the upload was successful, add the new training_run to the list to keep the names updated.
        self._training_run = tr_res.json()["id"]

        # upload model state. TODO figure out how this fits with multiple model states.
        self._upload_model_state(
            model=model,
            training_run_pk=self._training_run,
            dataset_pks=dataset_pks,
            sequence_num=0,
            final=True,
            model_manager=framework,
        )

    @staticmethod
    def _assemble_dataset(data: Dict[str, DataFrame]) -> DataFrame:
        if len(data) == 1:
            return next(iter(data.values()))
        elif len(data) == 2:
            # create dataframe from X and y and upload - will have one item in metadata, the output_col
            for key, val in data.items():
                if not (isinstance(val, DataFrame) or isinstance(val, Series)):
                    data[key] = DataFrame(val)
            return pd.concat([x for x in data.values()], axis=1)
        else:
            raise AssertionError(
                "Output doesn't match the requirements. Please review the documentation."
            )

    def _init_project(self, project_name) -> None:
        """
        Initialises the project for the object. If the project does not exist on the server it will be created.

        :param project_name: The name of the project

        :return: None
        """
        self._project = self._get_project(project_name)
        if self._project is None:
            proj_res = self._create_project(project_name=project_name)
            try:
                self._project = proj_res.json()["id"]
            except KeyError:
                print(f"There was an issue: {proj_res.text}")
                resp = self._transmission.get(
                    url_path="/collection/projects", query_params={"name": project_name}
                )
                resp = handle_response(
                    resp, 200, f"There was an issue getting the project: {resp.reason}"
                )
                self._project = resp.json()[0]["id"]

    def _get_project(self, project_name: str) -> Any:
        """
        Checks if a project exists on the server. If it does not it will return None otherwise the id of the project.

        :param project_name: str The name of the project.

        :return: int | None The id of the project else None.
        """
        project_res = self._transmission.get(
            url_path="/collection/projects",
            query_params={"name": project_name, "organization": self._organization},
        )
        handle_response(
            project_res, 200, f"There was an issue getting the projects: {project_res.reason}"
        )
        if len(project_res.json()) == 0:
            return None
        return project_res.json()[0]["id"]

    def _create_project(self, project_name: str, description: str = "Please add a description.."):
        """
        Creates a new project.

        :param project_name: str The name of the project, must be unique within your Organisation.

        :param description: str Optional The description of the project. This has a default value that can be changed
            at a later date

        :return: Response The response from the server.

        :raises ValueError if the response status is not 201.
        """
        res = self._transmission.send_json(
            url_path="/collection/projects",
            obj={
                "name": project_name,
                "description": description,
                "organization": self._organization,
            },
            query_params={"organization": self._organization},
        )
        return handle_response(
            res, expected=201, msg=f"There was an issue creating the project: {res.reason}"
        )

    def _set_model(self, model_name: str, framework: ModelManagers) -> int:
        """
        Set the model for this session.
        Checks if it has already been uploaded. If not it will upload it.

        :param model_name: The name for the architecture/algorithm. eg. "GradientBoostedMachine" or "3-layer CNN".

        :return: int The model id.

        :raises: ValueError - if the framework is not one of the supported frameworks or if there is an issue uploading
         the model.
        """
        res = handle_response(
            self._transmission.get(
                url_path="/collection/models",
                query_params={
                    "organization": self._organization,
                    "project": self._project,
                    "name": model_name,
                    "framework": framework.name,
                },
            ),
            expected=200,
            msg="There was an issue getting the model list",
        )
        models = res.json()
        if len(models) == 1:
            return models[0]["id"]
        # if we got here that means that the model has not been uploaded yet. So we upload it.
        res = self._upload_model(model_name=model_name, framework=framework)
        try:
            model_pk = res.json()["id"]
        except KeyError:
            resp = handle_response(
                self._transmission.get(
                    url_path="/collection/models",
                    query_params={
                        "organization": self._organization,
                        "project": self._project,
                        "name": model_name,
                        "framework": framework.name,
                    },
                ),
                expected=200,
                msg="There was an issue getting the model list",
            )
            model_pk = resp.json()[0]["id"]
        return model_pk

    @staticmethod
    def _ensure_required_metadata(metadata: Dict, defaults_spec: Dict) -> Dict:
        """
        Ensures that required metadata that can be specified by the user are filled.
        @param metadata: The metadata dict
        @param defaults_spec:
        @return: metadata
        """
        for required_key, default in defaults_spec.items():
            try:
                if metadata[required_key] is None:
                    metadata[required_key] = default
            except KeyError:
                metadata[required_key] = default
        return metadata

    @staticmethod
    def _add_required_metadata(metadata: Dict, required_spec: Dict) -> Dict:
        """
        Adds required - non user specified fields to the metadata
        @param metadata: The metadata dict
        @param required_spec:
        @return: metadata
        """
        for required_key, default in required_spec.items():
            metadata[required_key] = default
        return metadata

    def _upload_dataset(
        self,
        dataset: DataFrame,
        dataset_name: str,
        metadata: Dict,
        parent_hash: Union[int, None],
        transformation: Union[DatasetTransformation, None],
    ):
        # upload a dataset - only works for a single transformation.
        if not os.path.exists(self._cache_dir):
            os.makedirs(self._cache_dir)

        # TODO refactor to make multithreading safe.
        dataset_file_path = save_dataset(
            dataset, dataset_name, os.path.join(self._cache_dir, uuid.uuid4().__str__())
        )
        dset_pk = dataset_hash(dataset, self._project)

        response = post_dataset(
            transmission=self._transmission,
            dataset_file_path=dataset_file_path,
            project_pk=self._project,
            organization_pk=self._organization,
            name=dataset_name,
            metadata=metadata,
            dataset_pk=dset_pk,
            parent_dataset_hash=str(parent_hash) if parent_hash is not None else None,
        )
        handle_response(
            response, 201, f"There was some issue uploading the dataset: {response.reason}"
        )
        # tidy up files.
        os.remove(dataset_file_path)

        # upload the transformations
        if response.status_code == 201:
            if transformation is not None:
                # upload transformations.
                self._upload_transformation(
                    transformation=transformation,
                    dataset_pk=dset_pk,
                )

    def _upload_model(self, model_name: str, framework: ModelManagers):
        """

        :param model_name:
        :param framework: instance of seclea_ai.Frameworks
        :return:
        """
        res = self._transmission.send_json(
            url_path="/collection/models",
            obj={
                "organization": self._organization,
                "project": self._project,
                "name": model_name,
                "framework": framework.name,
            },
            query_params={"organization": self._organization, "project": self._project},
        )
        return handle_response(
            res, expected=201, msg=f"There was an issue uploading the model: {res.reason}"
        )

    def _upload_training_run(
        self, training_run_name: str, model_pk: int, dataset_pks: List[str], params: Dict
    ):
        """

        :param training_run_name: eg. "Training Run 0"
        :param params: Dict The hyper parameters of the model - can auto extract?
        :return:
        """
        if self._project is None:
            raise Exception("You need to create a project before uploading a training run")
        res = self._transmission.send_json(
            url_path="/collection/training-runs",
            obj={
                "organization": self._organization,
                "project": self._project,
                "datasets": dataset_pks,
                "model": model_pk,
                "name": training_run_name,
                "params": params,
            },
            query_params={"organization": self._organization, "project": self._project},
        )
        return handle_response(
            res, expected=201, msg=f"There was an issue uploading the training run: {res.reason}"
        )

    def _upload_model_state(
        self,
        model,
        training_run_pk: int,
        dataset_pks: List[int],
        sequence_num: int,
        final: bool,
        model_manager: ModelManagers,
    ):
        os.makedirs(
            os.path.join(self._cache_dir, str(training_run_pk)),
            exist_ok=True,
        )
        model_data = serialize(model, model_manager)
        file_name = f"data-{dataset_pks[0]}-model-{sequence_num}"
        save_path = os.path.join(Path.home(), f".seclea/{self._project_name}/{training_run_pk}")
        save_path = save_object(
            model_data, file_name, save_path, compression=CompressionFactory.ZSTD
        )

        res = post_model_state(
            self._transmission,
            save_path,
            self._organization,
            self._project,
            str(training_run_pk),
            sequence_num,
            final,
        )

        # tidy up files.
        os.remove(save_path)

        res = handle_response(
            res, expected=201, msg=f"There was an issue uploading a model state: {res.reason}"
        )
        return res

    def _upload_transformation(self, transformation: DatasetTransformation, dataset_pk):
        idx = 0
        trans_kwargs = {**transformation.data_kwargs, **transformation.kwargs}
        data = {
            "name": transformation.func.__name__,
            "code_raw": inspect.getsource(transformation.func),
            "code_encoded": encode_func(transformation.func, [], trans_kwargs),
            "order": idx,
            "dataset": dataset_pk,
        }
        res = self._transmission.send_json(
            url_path="/collection/dataset-transformations",
            obj=data,
            query_params={"organization": self._organization, "project": self._project},
        )
        res = handle_response(
            res,
            expected=201,
            msg=f"There was an issue uploading the transformations on transformation {idx} with name {transformation.func.__name__}: {res.reason}",
        )
        return res

    def _load_transformations(self, training_run_pk: int):
        """
        Expects a list of code_encoded as set by upload_transformations.
        TODO replace or remove
        """
        res = self._transmission.get(
            url_path="/collection/dataset-transformations",
            query_params={"training_run": training_run_pk},
        )
        res = handle_response(
            res, expected=200, msg=f"There was an issue loading the transformations: {res.reason}"
        )
        transformations = list(map(lambda x: x["code_encoded"], res.json()))
        return list(map(decode_func, transformations))

    @staticmethod
    def _aggregate_dataset(datasets: List[str], index) -> DataFrame:
        """
        Aggregates a list of dataset paths into a single file for upload.
        NOTE the files must be split by row and have the same format otherwise this will fail or cause unexpected format
        issues later.
        :param datasets:
        :return:
        """
        loaded_datasets = [pd.read_csv(dset, index_col=index) for dset in datasets]
        aggregated = pd.concat(loaded_datasets, axis=0)
        return aggregated

    @staticmethod
    def _get_framework(model) -> ModelManagers:
        module = model.__class__.__module__
        # order is important as xgboost and lightgbm contain sklearn compliant packages.
        # TODO check if we can treat them as sklearn but for now we avoid that issue by doing sklearn last.
        if "xgboost" in module:
            return ModelManagers.XGBOOST
        elif "lightgbm" in module:
            return ModelManagers.LIGHTGBM
        # TODO improve to exclude other keras backends...
        elif "tensorflow" in module or "keras" in module:
            return ModelManagers.TENSORFLOW
        elif "sklearn" in module:
            return ModelManagers.SKLEARN
        else:
            return ModelManagers.NOT_IMPORTED

    @staticmethod
    def _get_dataset_type(dataset: DataFrame) -> str:
        if not np.issubdtype(dataset.index.dtype, np.integer):
            try:
                pd.to_datetime(dataset.index.values)
            except (ParserError, ValueError):  # Can't cnvrt some
                return "tabular"
            return "time_series"
        return "tabular"
